from bisect import bisect_right
from typing import List, Optional, Tuple, Union
from warnings import warn

from PySide6.QtCore import QMimeData, QPoint, QSize, Signal, SignalInstance
from PySide6.QtGui import (
    QDragEnterEvent,
    QDragMoveEvent,
    QMouseEvent,
    QPainter,
    QPaintEvent,
    QPixmap,
    Qt,
    QWheelEvent,
)
from PySide6.QtWidgets import QSizePolicy, QToolTip, QWidget

from foundry.core.Position import Position
from foundry.game.gfx.drawable.Block import Block
from foundry.game.gfx.objects.EnemyItem import EnemyObject
from foundry.game.gfx.objects.LevelObject import LevelObject
from foundry.game.gfx.objects.ObjectLike import (
    EXPANDS_BOTH,
    EXPANDS_HORIZ,
    EXPANDS_VERT,
)
from foundry.game.level.Level import Level
from foundry.game.level.LevelRef import LevelRef
from foundry.game.level.WorldMap import WorldMap
from foundry.gui.ContextMenu import ContextMenu
from foundry.gui.LevelDrawer import LevelDrawer
from foundry.gui.SelectionSquare import SelectionSquare
from foundry.gui.settings import RESIZE_LEFT_CLICK, RESIZE_RIGHT_CLICK, SETTINGS

HIGHEST_ZOOM_LEVEL = 8  # on linux, at least
LOWEST_ZOOM_LEVEL = 1 / 16  # on linux, but makes sense with 16x16 blocks

# mouse modes

MODE_FREE = 0
MODE_DRAG = 1
MODE_RESIZE_HORIZ = 2
MODE_RESIZE_VERT = 4
MODE_RESIZE_DIAG = MODE_RESIZE_HORIZ | MODE_RESIZE_VERT
RESIZE_MODES = [MODE_RESIZE_HORIZ, MODE_RESIZE_VERT, MODE_RESIZE_DIAG]


def undoable(func):
    def wrapped(self, *args):
        func(self, *args)
        self.level_ref.save_level_state()

    return wrapped


class LevelView(QWidget):
    objects_selected: SignalInstance = Signal(object)  # type: ignore
    object_created: SignalInstance = Signal(object)  # type: ignore

    def __init__(self, parent: Optional[QWidget], level: LevelRef, context_menu: ContextMenu):
        super(LevelView, self).__init__(parent)

        self.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.setMouseTracking(True)
        self.setAcceptDrops(True)

        self.level_ref: LevelRef = level
        self.level_ref.data_changed.connect(self.update)

        self.context_menu = context_menu

        self.level_drawer = LevelDrawer()

        self.draw_grid = SETTINGS["draw_grid"]
        self.draw_jumps = SETTINGS["draw_jumps"]
        self.draw_expansions = SETTINGS["draw_expansion"]
        self.draw_mario = SETTINGS["draw_mario"]
        self.transparency = SETTINGS["block_transparency"]
        self.draw_jumps_on_objects = SETTINGS["draw_jump_on_objects"]
        self.draw_items_in_blocks = SETTINGS["draw_items_in_blocks"]
        self.draw_invisible_items = SETTINGS["draw_invisible_items"]
        self.draw_autoscroll = SETTINGS["draw_autoscroll"]

        self.zoom = 1
        self.block_length = Block.SIDE_LENGTH * self.zoom

        self.changed = False

        self.selection_square = SelectionSquare()

        self.mouse_mode = MODE_FREE

        self.last_mouse_position = 0, 0

        self.drag_start_point = 0, 0

        self.dragging_happened = True

        self.resize_mouse_start_x = 0
        self.resize_obj_start_point = 0, 0

        self.resizing_happened = False

        # dragged in from the object toolbar
        self.currently_dragged_object: Optional[Union[LevelObject, EnemyObject]] = None

        self.setWhatsThis(
            "<b>Level View</b><br/>"
            "This renders the level as it would appear in game plus additional information, that can be "
            "toggled in the View menu.<br/>"
            "It supports selecting multiple objects, moving, copy/pasting and resizing them using the "
            "mouse or the usual keyboard shortcuts.<br/>"
            "There are still occasional rendering errors, or small inconsistencies. If you find them, "
            "please report the kind of object (name or values in the SpinnerPanel) and the level or "
            "object set they appear in, in the discord and @Michael or on the github page under Help."
            "<br/><br/>"
            ""
            "If all else fails, click the play button up top to see your level in game in seconds."
        )

    @property
    def transparency(self):
        return self.level_drawer.transparency

    @transparency.setter
    def transparency(self, value):
        self.level_drawer.transparency = value

    @property
    def draw_grid(self):
        return self.level_drawer.draw_grid

    @draw_grid.setter
    def draw_grid(self, value):
        self.level_drawer.draw_grid = value

    @property
    def draw_jumps(self):
        return self.level_drawer.draw_jumps

    @draw_jumps.setter
    def draw_jumps(self, value):
        self.level_drawer.draw_jumps = value

    @property
    def draw_mario(self):
        return self.level_drawer.draw_mario

    @draw_mario.setter
    def draw_mario(self, value):
        self.level_drawer.draw_mario = value

    @property
    def draw_expansions(self):
        return self.level_drawer.draw_expansions

    @draw_expansions.setter
    def draw_expansions(self, value):
        self.level_drawer.draw_expansions = value

    @property
    def draw_jumps_on_objects(self):
        return self.level_drawer.draw_jumps_on_objects

    @draw_jumps_on_objects.setter
    def draw_jumps_on_objects(self, value):
        self.level_drawer.draw_jumps_on_objects = value

    @property
    def draw_items_in_blocks(self):
        return self.level_drawer.draw_items_in_blocks

    @draw_items_in_blocks.setter
    def draw_items_in_blocks(self, value):
        self.level_drawer.draw_items_in_blocks = value

    @property
    def draw_invisible_items(self):
        return self.level_drawer.draw_invisible_items

    @draw_invisible_items.setter
    def draw_invisible_items(self, value):
        self.level_drawer.draw_invisible_items = value

    @property
    def draw_autoscroll(self):
        return self.level_drawer.draw_autoscroll

    @draw_autoscroll.setter
    def draw_autoscroll(self, value):
        self.level_drawer.draw_autoscroll = value

    def mousePressEvent(self, event: QMouseEvent):
        pressed_button = event.button()

        if pressed_button == Qt.LeftButton:
            self._on_left_mouse_button_down(event)
        elif pressed_button == Qt.RightButton:
            self._on_right_mouse_button_down(event)
        else:
            return super(LevelView, self).mousePressEvent(event)

    def mouseMoveEvent(self, event: QMouseEvent):
        if self.mouse_mode == MODE_DRAG:
            self.setCursor(Qt.ClosedHandCursor)
            self._dragging(event)

        elif self.mouse_mode in RESIZE_MODES:
            previously_selected_objects = self.level_ref.selected_objects

            self._resizing(event)

            self.level_ref.selected_objects = previously_selected_objects

        elif self.selection_square.active:
            self._set_selection_end(event.position().toPoint())

        elif SETTINGS["resize_mode"] == RESIZE_LEFT_CLICK:
            self._set_cursor_for_position(event)

        x, y = event.position().toPoint().toTuple()

        object_under_cursor = self.object_at(x, y)

        if SETTINGS["object_tooltip_enabled"] and object_under_cursor is not None:
            self.setToolTip(str(object_under_cursor))
        else:
            self.setToolTip("")
            QToolTip.hideText()

        return super(LevelView, self).mouseMoveEvent(event)

    def _set_cursor_for_position(self, event: QMouseEvent):
        level_object = self.object_at(*event.position().toTuple())

        if level_object is not None:
            if isinstance(level_object, LevelObject):
                is_resizable = not level_object.is_single_block
            else:
                is_resizable = False

            edges = self._cursor_on_edge_of_object(level_object, event.position().toPoint())

            if is_resizable and edges:
                if edges == Qt.RightEdge and level_object.expands() & EXPANDS_HORIZ:
                    cursor = Qt.SizeHorCursor
                elif edges == Qt.BottomEdge and level_object.expands() & EXPANDS_VERT:
                    cursor = Qt.SizeVerCursor
                elif (level_object.expands() & EXPANDS_BOTH) == EXPANDS_BOTH:
                    cursor = Qt.SizeFDiagCursor
                else:
                    return

                if self.mouse_mode not in RESIZE_MODES:
                    self.setCursor(cursor)

                return

        if self.mouse_mode not in RESIZE_MODES:
            self.setCursor(Qt.ArrowCursor)

    def _cursor_on_edge_of_object(
        self, level_object: Union[LevelObject, EnemyObject], pos: QPoint, edge_width: int = 4
    ):
        right = (level_object.get_rect().left() + level_object.get_rect().width()) * self.block_length
        bottom = (level_object.get_rect().top() + level_object.get_rect().height()) * self.block_length

        on_right_edge = pos.x() in range(right - edge_width, right)
        on_bottom_edge = pos.y() in range(bottom - edge_width, bottom)

        edges = 0

        if on_right_edge:
            edges |= Qt.RightEdge

        if on_bottom_edge:
            edges |= Qt.BottomEdge

        return edges

    def mouseReleaseEvent(self, event: QMouseEvent):
        released_button = event.button()

        if released_button == Qt.LeftButton:
            self._on_left_mouse_button_up(event)
        elif released_button == Qt.RightButton:
            self._on_right_mouse_button_up(event)
        else:
            super(LevelView, self).mouseReleaseEvent(event)

    def wheelEvent(self, event: QWheelEvent):
        if SETTINGS["object_scroll_enabled"]:
            x, y = event.position().toPoint().toTuple()

            obj_under_cursor = self.object_at(x, y)

            if obj_under_cursor is None:
                return False

            if isinstance(self.level_ref.level, WorldMap):
                return False

            # scrolling through the level could unintentionally change objects, if the cursor would wander onto them.
            # this is annoying (to me) so only change already selected objects
            if obj_under_cursor not in self.level_ref.selected_objects:
                return False

            self._change_object_on_mouse_wheel(event.position().toPoint(), event.angleDelta().y())

            return True
        else:
            super(LevelView, self).wheelEvent(event)
            return False

    @undoable
    def _change_object_on_mouse_wheel(self, cursor_position: QPoint, y_delta: int):
        x, y = cursor_position.x(), cursor_position.y()

        obj = self.object_at(x, y)

        if obj is None:
            return
        obj.increment_type() if y_delta > 0 else obj.decrement_type()
        obj.selected = True

    def sizeHint(self) -> QSize:
        if not self.level_ref:
            return super(LevelView, self).sizeHint()
        else:
            width, height = self.level_ref.level.size

            return QSize(width * self.block_length, height * self.block_length)

    def update(self):
        self.resize(self.sizeHint())

        super(LevelView, self).update()

    def _on_right_mouse_button_down(self, event: QMouseEvent):
        if self.mouse_mode == MODE_DRAG:
            return

        x, y = event.position().toPoint().toTuple()
        level_x, level_y = self._to_level_point(x, y)

        self.last_mouse_position = level_x, level_y

        if self._select_objects_on_click(event) and SETTINGS["resize_mode"] == RESIZE_RIGHT_CLICK:
            self._try_start_resize(MODE_RESIZE_DIAG, event)

    def _try_start_resize(self, resize_mode: int, event: QMouseEvent):
        if resize_mode not in RESIZE_MODES:
            return

        x, y = event.position().toPoint().toTuple()
        level_x, level_y = self._to_level_point(x, y)

        self.mouse_mode = resize_mode

        self.resize_mouse_start_x = level_x

        obj = self.object_at(x, y)

        if obj is not None:
            self.resize_obj_start_point = obj.position.x, obj.position.y

    def _resizing(self, event: QMouseEvent):
        self.resizing_happened = True

        if isinstance(self.level_ref.level, WorldMap):
            return

        x, y = event.position().toPoint().toTuple()

        level_x, level_y = self._to_level_point(x, y)

        dx = dy = 0

        if self.mouse_mode & MODE_RESIZE_HORIZ:
            dx = level_x - self.resize_obj_start_point[0]

        if self.mouse_mode & MODE_RESIZE_VERT:
            dy = level_y - self.resize_obj_start_point[1]

        self.last_mouse_position = level_x, level_y

        selected_objects = self.get_selected_objects()

        for obj in selected_objects:
            obj.resize_by(dx, dy)

            self.level_ref.level.changed = True

        self.update()

    def _on_right_mouse_button_up(self, event: QMouseEvent):
        if self.resizing_happened:
            self._on_resize_happened_mouse_up(event)
        else:
            if self.get_selected_objects():
                menu = self.context_menu.as_object_menu()
            else:
                menu = self.context_menu.as_background_menu()

            self.context_menu.set_position(event.position().toPoint())

            menu_pos = self.mapToGlobal(event.position().toPoint())

            menu.popup(menu_pos)

        self.resizing_happened = False
        self.mouse_mode = MODE_FREE
        self.setCursor(Qt.ArrowCursor)

    def _stop_resize(self):
        self.level_ref.save_level_state()
        self.resizing_happened = False
        self.mouse_mode = MODE_FREE
        self.setCursor(Qt.ArrowCursor)

    def _on_left_mouse_button_down(self, event: QMouseEvent):
        if self._select_objects_on_click(event):
            x, y = event.position().toPoint().toTuple()

            obj = self.object_at(x, y)

            if obj is not None:
                edge = self._cursor_on_edge_of_object(obj, event.position().toPoint())

                if SETTINGS["resize_mode"] == RESIZE_LEFT_CLICK and edge:

                    self._try_start_resize(self._resize_mode_from_edge(edge), event)
                else:
                    self.drag_start_point = obj.position.x, obj.position.y
        else:
            self._start_selection_square(event.position().toPoint())

    def _on_resize_happened_mouse_up(self, event: QMouseEvent):
        x, y = event.pos().x(), event.pos().y()

        resize_end_x, _ = self._to_level_point(x, y)
        if self.resize_mouse_start_x != resize_end_x:
            self._stop_resize()

    @staticmethod
    def _resize_mode_from_edge(edge: int):
        mode = 0

        if edge & Qt.RightEdge:
            mode |= MODE_RESIZE_HORIZ

        if edge & Qt.BottomEdge:
            mode |= MODE_RESIZE_VERT

        return mode

    def _dragging(self, event: QMouseEvent):
        self.dragging_happened = True

        x, y = event.position().toPoint().toTuple()

        level_x, level_y = self._to_level_point(x, y)

        dx = level_x - self.last_mouse_position[0]
        dy = level_y - self.last_mouse_position[1]

        self.last_mouse_position = level_x, level_y

        selected_objects = self.get_selected_objects()

        for obj in selected_objects:
            obj.move_by(dx, dy)

            self.level_ref.level.changed = True

        self.update()

    def _on_left_mouse_button_up(self, event: QMouseEvent):
        if self.resizing_happened:
            self._on_resize_happened_mouse_up(event)
        elif self.mouse_mode == MODE_DRAG and self.dragging_happened:
            x, y = event.position().toPoint().toTuple()

            obj = self.object_at(x, y)

            if obj is not None:
                drag_end_point = obj.position.x, obj.position.y

                if self.drag_start_point != drag_end_point:
                    self._stop_drag()
                else:
                    self.dragging_happened = False
        else:
            self._stop_selection_square()

        self.mouse_mode = MODE_FREE
        self.setCursor(Qt.ArrowCursor)

    def _stop_drag(self):
        if self.dragging_happened:
            self.level_ref.save_level_state()

        self.dragging_happened = False

    def _select_objects_on_click(self, event: QMouseEvent) -> bool:
        x, y = event.position().toPoint().toTuple()
        level_x, level_y = self._to_level_point(x, y)

        self.last_mouse_position = level_x, level_y

        clicked_object = self.object_at(x, y)
        clicked_on_background = clicked_object is None

        if clicked_on_background and not event.modifiers() & Qt.ShiftModifier:
            self._select_object(None)
        else:
            self.mouse_mode = MODE_DRAG

            selected_objects = self.get_selected_objects()
            nothing_selected = not selected_objects
            if nothing_selected or (
                not event.modifiers() & Qt.ShiftModifier and not event.modifiers() & Qt.ControlModifier
            ):
                self._select_object([clicked_object])
            else:
                self._select_object([clicked_object] + selected_objects)  # type: ignore

        return not clicked_on_background

    def _set_zoom(self, zoom):
        if not (LOWEST_ZOOM_LEVEL <= zoom <= HIGHEST_ZOOM_LEVEL):
            return

        self.zoom = zoom
        self.block_length = int(Block.SIDE_LENGTH * self.zoom)

        self.update()

    def zoom_out(self):
        self._set_zoom(self.zoom / 2)

    def zoom_in(self):
        self._set_zoom(self.zoom * 2)

    def _start_selection_square(self, position):
        self.selection_square.start(position)

    def _set_selection_end(self, position):
        if not self.selection_square.is_active():
            return

        self.selection_square.set_current_end(position)

        sel_rect = self.selection_square.get_adjusted_rect(self.block_length, self.block_length)

        touched_objects = [obj for obj in self.level_ref.level.get_all_objects() if sel_rect.intersects(obj.get_rect())]

        if touched_objects != self.level_ref.selected_objects:
            self._set_selected_objects(touched_objects)

        self.update()

    def _stop_selection_square(self):
        self.selection_square.stop()

        self.update()

    def select_all(self):
        self.select_objects(self.level_ref.level.get_all_objects())

    def _select_object(self, objects: Optional[list]):
        if objects is not None:
            self.select_objects(objects)
            self.objects_selected.emit(objects)
        else:
            self.select_objects([])
            self.objects_selected.emit([])

    def select_objects(self, objects):
        self._set_selected_objects(objects)

        self.update()

    def _set_selected_objects(self, objects):
        if self.level_ref.selected_objects == objects:
            return

        self.level_ref.selected_objects = objects

    def get_selected_objects(self) -> List[Union[LevelObject, EnemyObject]]:
        return self.level_ref.selected_objects

    def remove_selected_objects(self):
        for obj in self.level_ref.selected_objects:
            self.level_ref.level.remove_object(obj)

    def scroll_to_objects(self, objects: List[LevelObject]):
        if not objects:
            return

        min_x = min([obj.position.x for obj in objects]) * self.block_length
        min_y = min([obj.position.y for obj in objects]) * self.block_length

        self.parent().parent().ensureVisible(min_x, min_y)

    def level_safe_to_save(self) -> Tuple[bool, str, str]:
        is_safe = True
        reason = ""
        additional_info = ""

        if self.level_ref.level.too_many_level_objects():
            level = self._cuts_into_other_objects()

            is_safe = False
            reason = "Too many level objects."

            if level:
                additional_info = f"Would overwrite data of '{level}'."
            else:
                additional_info = (
                    "It wouldn't overwrite another level, " "but it might still overwrite other important data."
                )

        elif self.level_ref.level.too_many_enemies_or_items():
            level = self._cuts_into_other_enemies()

            is_safe = False
            reason = "Too many enemies or items."

            if level:
                additional_info = f"Would probably overwrite enemy/item data of '{level}'."
            else:
                additional_info = (
                    "It wouldn't overwrite enemy/item data of another level, "
                    "but it might still overwrite other important data."
                )

        return is_safe, reason, additional_info

    def _cuts_into_other_enemies(self) -> str:
        if self.level_ref is None:
            raise ValueError("Level is None")

        enemies_end = self.level_ref.level.enemies_end

        levels_by_enemy_offset = sorted(Level.offsets, key=lambda level: level.enemy_pointer)

        level_index = bisect_right([level.enemy_pointer for level in levels_by_enemy_offset], enemies_end) - 1

        found_level = levels_by_enemy_offset[level_index]

        if found_level.enemy_pointer == self.level_ref.level.enemy_offset:
            return ""
        else:
            return (
                f"World {found_level.display_information.locations[0].world} - {found_level.display_information.name}"
            )

    def _cuts_into_other_objects(self) -> str:
        if self.level_ref is None:
            raise ValueError("Level is None")

        end_of_level_objects = self.level_ref.level.objects_end

        level_index = (
            bisect_right(
                [level.generator_pointer - Level.HEADER_LENGTH for level in Level.sorted_offsets], end_of_level_objects
            )
            - 1
        )

        found_level = Level.sorted_offsets[level_index]

        if found_level.generator_pointer == self.level_ref.level.object_offset:
            return ""
        else:
            return (
                f"World {found_level.display_information.locations[0].world} - {found_level.display_information.name}"
            )

    def add_jump(self):
        self.level_ref.level.add_jump()

    def object_at(self, x: int, y: int) -> Optional[Union[LevelObject, EnemyObject]]:
        """
        Returns an enemy or level object at the position. The x and y is relative to the View (for example, when you
        receive a mouse event) and will be converted into level coordinates internally.

        :param int x: X position on the View, where the object is queried at.
        :param int y: Y position on the View, where the object is queried at.

        :return: An enemy/level object, or None, if none is at the position.
        """
        level_x, level_y = self._to_level_point(x, y)

        return self.level_ref.level.object_at(level_x, level_y)

    def _to_level_point(self, screen_x: int, screen_y: int) -> Tuple[int, int]:
        level_x = screen_x // self.block_length
        level_y = screen_y // self.block_length

        return level_x, level_y

    def create_object_at(self, x: int, y: int, domain: int = 0, object_index: int = 0):
        level_x, level_y = self._to_level_point(x, y)

        self.level_ref.level.create_object_at(int(level_x), int(level_y), domain, object_index)

        self.update()

    def create_enemy_at(self, x: int, y: int):
        level_x, level_y = self._to_level_point(x, y)

        self.level_ref.level.create_enemy_at(level_x, level_y)

    def add_object(self, domain: int, obj_index: int, x: int, y: int, length: int, index: int = -1):
        level_x, level_y = self._to_level_point(x, y)

        self.level_ref.level.add_object(domain, obj_index, level_x, level_y, length, index)

    def add_enemy(self, enemy_index: int, x: int, y: int, index: int):
        level_x, level_y = self._to_level_point(x, y)

        self.level_ref.level.add_enemy(enemy_index, level_x, level_y, index)

    def replace_object(self, obj: LevelObject, domain: int, obj_index: int, length: Optional[int]):
        self.remove_object(obj)

        new_obj = self.level_ref.level.add_object(
            domain, obj_index, obj.position.x, obj.position.y, length, obj.index_in_level
        )
        new_obj.selected = obj.selected

    def replace_enemy(self, old_enemy: EnemyObject, enemy_index: int):
        index_in_level = self.level_ref.level.index_of(old_enemy)

        self.remove_object(old_enemy)

        new_enemy = self.level_ref.level.add_enemy(
            enemy_index, old_enemy.position.x, old_enemy.position.y, index_in_level
        )

        new_enemy.selected = old_enemy.selected

    def remove_object(self, obj):
        self.level_ref.level.remove_object(obj)

    def remove_jump(self, index: int):
        del self.level_ref.level.jumps[index]

        self.update()

    def paste_objects_at(
        self,
        paste_data: Tuple[List[Union[LevelObject, EnemyObject]], Tuple[int, int]],
        x: Optional[int] = None,
        y: Optional[int] = None,
    ):
        if x is None or y is None:
            level_x, level_y = self.last_mouse_position
        else:
            level_x, level_y = self._to_level_point(x, y)

        objects, origin = paste_data

        ori_x, ori_y = origin

        pasted_objects = []

        for obj in objects:
            offset_x, offset_y = obj.position.x - ori_x, obj.position.y - ori_y

            try:
                pasted_objects.append(self.level_ref.level.paste_object_at(level_x + offset_x, level_y + offset_y, obj))
            except ValueError:
                warn("Tried pasting outside of level.", RuntimeWarning)

        self.select_objects(pasted_objects)

    def get_object_names(self):
        return self.level_ref.level.get_object_names()

    def make_screenshot(self) -> QPixmap:
        assert self.level_ref is not None

        return self.grab()

    def dragEnterEvent(self, event: QDragEnterEvent):
        if event.mimeData().hasFormat("application/level-object"):
            event.acceptProposedAction()

    def dragMoveEvent(self, event: QDragMoveEvent):
        x, y = self._to_level_point(*event.position().toPoint().toTuple())

        level_object = self._object_from_mime_data(event.mimeData())

        level_object.position = Position(x, y)

        self.currently_dragged_object = level_object

        self.repaint()

    def dragLeaveEvent(self, event):
        self.currently_dragged_object = None

        self.repaint()

    @undoable
    def dropEvent(self, event):
        x, y = self._to_level_point(*event.position().toPoint().toTuple())

        level_object = self._object_from_mime_data(event.mimeData())

        if isinstance(level_object, LevelObject):
            self.level_ref.level.add_object(level_object.domain, level_object.obj_index, x, y, None)
        else:
            self.level_ref.level.add_enemy(level_object.obj_index, x, y)

        event.accept()

        self.currently_dragged_object = None

        self.object_created.emit(level_object)
        self.level_ref.data_changed.emit()

    def _object_from_mime_data(self, mime_data: QMimeData) -> Union[LevelObject, EnemyObject]:
        object_type, *object_bytes = mime_data.data("application/level-object")

        if object_type == b"\x00":
            domain = int.from_bytes(object_bytes[0], "big") >> 5
            object_index = int.from_bytes(object_bytes[2], "big")

            return self.level_ref.level.object_factory.from_properties(domain, object_index, 0, 0, None, 999)
        else:
            enemy_id = int.from_bytes(object_bytes[0], "big")

            return self.level_ref.level.enemy_item_factory.from_properties(enemy_id, 0, 0)

    def paintEvent(self, event: QPaintEvent):
        painter = QPainter(self)

        if self.level_ref is None:
            return

        self.level_drawer.block_length = self.block_length

        self.level_drawer.draw(painter, self.level_ref.level)

        self.selection_square.draw(painter)

        if self.currently_dragged_object is not None:
            self.currently_dragged_object.draw(painter, self.block_length, self.transparency)



from PySide6.QtCore import QRect

from foundry.core.Position import Position, PositionProtocol
from foundry.game.Definitions import Definition
from foundry.game.gfx.objects.ObjectLike import ObjectLike

map_object_names = {
    0x00: "Mario Clear (Blue)",
    0x01: "Luigi Clear (Blue)",
    0x02: "Black Square",
    0x03: "Level 1",
    0x04: "Level 2",
    0x05: "Level 3",
    0x06: "Level 4",
    0x07: "Level 5",
    0x08: "Level 6",
    0x09: "Level 7",
    0x0A: "Level 8",
    0x0B: "Level 9",
    0x0C: "Level 10",
    0x0D: "Level 1 (Broken)",
    0x0E: "Level 2 (Broken)",
    0x0F: "Level 3 (Broken)",
    0x10: "Level 4 (Broken)",
    0x11: "Level 5 (Broken)",
    0x12: "Level 6 (Broken)",
    0x13: "Level 7 (Broken)",
    0x14: "Level 8 (Broken)",
    0x15: "Level 9 (Broken)",
    0x40: "Mario Clear (Orange)",
    0x41: "Luigi Clear (Orange)",
    0x42: "Desert Background",
    0x43: "Sand",
    0x44: "Path Upper Left",
    0x45: "Path Horizontal",
    0x46: "Path Vertical",
    0x47: "Path Upper Right",
    0x48: "Path Lower Left",
    0x49: "Path Horizontal 2",
    0x4A: "Path Lower Right",
    0x4B: "Pier",
    0x4C: "I's",
    0x4D: "Z's",
    0x4E: "? 1",
    0x4F: "? 2",
    0x50: "Mushroom House (Orange)",
    0x51: "Rock 1",
    0x52: "Rock 2",
    0x53: "Rock 3",
    0x54: "Key Door 1",
    0x55: "Star",
    0x56: "Key Door 2",
    0x57: "Miniature Path Lower Right",
    0x58: "Miniature Path Lower Left",
    0x59: "Miniature Path Horizontal",
    0x5A: "Miniature Tower",
    0x5B: "Miniature Path Point Horizontal",
    0x5C: "Miniature Path Lower Left 2",
    0x5D: "Miniature Cacti",
    0x5E: "Miniature Cacti 2",
    0x5F: "Tower",
    0x60: "Fortress Ruins",
    0x61: "Bowsers Castle Wall Tower",
    0x62: "Bowsers Castle Wall Side",
    0x63: "Bowsers Castle Wall Top 1",
    0x64: "Bowsers Castle Wall",
    0x65: "Bowsers Castle Wall Top 2",
    0x66: "Path Upper Right 2",
    0x67: "Fortress",
    0x68: "Quicksand",
    0x69: "Pyramid",
    0x6A: "Barracks",
    0x80: "Mario Clear (Green)",
    0x81: "Luigi Clear (Green)",
    0x82: "Water Three-Way Up",
    0x83: "Water Three-Way Down",
    # TODO continue
    0xB1: "Switchable Bridge Vertical",
    0xB2: "Switchable Bridge Horizontal",
    0xB3: "Round Bridge",
    0xB4: "Bushes",
    # TODO continue
    0xBB: "Palm Tree",
    0xBC: "Pipe",
    0xBD: "Fire Flower",
    0xBE: "Piranha Plant",
    0xBF: "Pond",
    0xC0: "Mario Clear (Red)",
    0xC1: "Luigi Clear (Red)",
    0xC2: "Cloud Upper Left",
    0xC3: "Cloud Top Left",
    0xC4: "Cloud Top Right",
    0xC5: "Cloud Upper Right",
    0xC6: "? 3",
    0xC7: "? 4",
    0xC8: "End Castle Top",
    0xC9: "End Castle Bottom",
    0xCA: "Bowsers Lair Top Left",
    0xCB: "Bowsers Lair Top Right",
    0xCC: "Bowsers Lair Bottom Left",
    0xCD: "Bowsers Lair Bottom Right",
    0xCE: "Cloud Left 1",
    0xCF: "? 5",
    0xD0: "Cloud Diagnoal",
    0xD1: "Flame",
    0xD2: "Cloud Left 2",
    0xD3: "Cloud Bottom",
    0xD4: "Cloud Lower Right",
    0xD5: "I's 2",
    0xD6: "Red Background ?",
    0xD7: "Desert Background 2 ?",
    0xD8: "Black Square",
    0xD9: "Path Upper Left 2",
    0xDA: "Path Horizontal 3",
    0xDB: "Path Vertical 2",
    0xDC: "Path Upper Right 2",
    0xDD: "Path Lower Left 2",
    0xDE: "Path Lower Right 2",
    0xDF: "Tower 2",
    0xE0: "Mushroom House 2",
    0xE1: "Mushroom",
    0xE2: "Skull",
    0xE3: "Fortress Ruins 2",
    0xE4: "Key Door 3",
    0xE5: "Start Field",
    0xE6: "Hand Field",
    0xE7: "? 6",
    0xE8: "Spade Bonus",
    0xE9: "Star 2",
    0xEA: "Rock Alternative",
    0xEB: "Fortress 2",
}


class MapObject(ObjectLike):
    def __init__(self, block, x, y):
        self.x_position = x
        self.y_position = y

        self.block = block

        self.rect = QRect(self.x_position, self.y_position, 1, 1)

        if self.block.index in map_object_names:
            self.name = map_object_names[self.block.index]
        else:
            self.name = str(hex(self.block.index))

        self.selected = False

    @property
    def definition(self) -> Definition:
        return Definition(description="Map Object", warnings=[])

    @property
    def position(self) -> PositionProtocol:
        return Position(self.x_position, self.y_position)

    @position.setter
    def position(self, position: PositionProtocol):
        self.x_position, self.y_position = position.x, position.y
        self.rect = QRect(position.x, position.y, 1, 1)

    def render(self):
        pass

    def draw(self, dc, block_length, _=None):
        self.block.draw(
            dc,
            self.x_position * block_length,
            self.y_position * block_length,
            block_length=block_length,
            selected=self.selected,
            transparent=False,
        )

    def get_status_info(self):
        return ("x", self.x_position), ("y", self.y_position), ("Block Type", self.name)

    def to_bytes(self):
        return self.block.index

    def move_by(self, dx, dy):
        self.position = Position(self.x_position + dx, self.y_position + dy)

    def resize_to(self, x, y):
        return

    def resize_by(self, dx, dy):
        return

    def point_in(self, x, y):
        return self.rect.contains(x, y)

    def change_type(self, new_type):
        pass

    def get_rect(self):
        return self.rect

    def __contains__(self, point):
        pass



from typing import Generic, Optional, TypeVar

from PySide6.QtCore import QMimeData, QSize, Qt, Signal, SignalInstance
from PySide6.QtGui import QDrag, QMouseEvent, QPainter, QPaintEvent
from PySide6.QtWidgets import QLabel, QSizePolicy, QVBoxLayout, QWidget

from foundry.game.gfx.objects.EnemyItem import EnemyObject
from foundry.game.gfx.objects.LevelObject import LevelObject
from foundry.game.gfx.objects.LevelObjectFactory import LevelObjectFactory
from foundry.gui.util import clear_layout

T = TypeVar("T", LevelObject, EnemyObject)


def get_minimal_icon_object(level_object: T) -> T:
    """
    Returns the object with a length, so that every block is rendered. E. g. clouds with length 0, don't have a face.
    """

    if not isinstance(level_object, LevelObject):
        return level_object

    factory = LevelObjectFactory(
        level_object.object_set.number,
        level_object.graphics_set,
        0,
        [],
        vertical_level=False,
        size_minimal=True,
    )

    if level_object.obj_index >= 0x1F:
        obj = factory.from_properties(
            domain=level_object.domain,
            object_index=(level_object.obj_index // 0x10) * 0x10,
            x=0,
            y=0,
            length=None,
            index=0,
        )
    else:
        obj = factory.from_properties(
            domain=level_object.domain, object_index=level_object.obj_index, x=0, y=0, length=None, index=0
        )
    if not isinstance(obj, LevelObject):
        return level_object

    obj.ground_level = 3

    while any(block not in obj.rendered_blocks for block in obj.blocks) and obj.length < 0x10:
        obj.length += 1

        if obj.is_4byte:
            obj.secondary_length += 1

        obj.render()

    return obj


class ObjectIcon(QWidget, Generic[T]):
    def __init__(
        self,
        parent: Optional[QWidget],
        item: T,
        background_color: bool = True,
    ):
        super().__init__(parent)

        self.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self._item = item
        self.setToolTip(self.item.name)
        self._background_color = background_color

    @property
    def item(self) -> T:
        return self._item

    @item.setter
    def item(self, item: T):
        self._item = item
        self.setToolTip(item.name)
        self.update()

    @property
    def background_color(self) -> bool:
        return self._background_color

    @background_color.setter
    def background_color(self, background_color: bool):
        if self._background_color != background_color:
            self._background_color = background_color
            self.update()

    def sizeHint(self):
        return QSize(32, 32)

    def paintEvent(self, event: QPaintEvent):
        painter = QPainter(self)

        if self.background_color:
            painter.fillRect(event.rect(), self.item.palette_group.background_color)

        scaled_image = get_minimal_icon_object(self.item).as_image().scaled(self.size(), aspectMode=Qt.KeepAspectRatio)

        x = (self.width() - scaled_image.width()) // 2
        y = (self.height() - scaled_image.height()) // 2

        painter.drawImage(x, y, scaled_image)

        return super(ObjectIcon, self).paintEvent(event)


class ObjectButton(ObjectIcon):
    selected: SignalInstance = Signal()  # type: ignore
    object_created: SignalInstance = Signal()  # type: ignore

    def mouseMoveEvent(self, event: QMouseEvent):
        if not (event.buttons() & Qt.LeftButton):
            return super().mouseMoveEvent(event)

        drag_event = QDrag(self)
        mime_data = QMimeData()
        object_bytes = bytearray()
        object_bytes.append(0 if isinstance(self.item, LevelObject) else 1)
        object_bytes.extend(self.item.to_bytes())
        mime_data.setData("application/level-object", object_bytes)
        drag_event.setMimeData(mime_data)

        if drag_event.exec_() == Qt.MoveAction:
            self.object_created.emit()

    def mouseReleaseEvent(self, event: QMouseEvent):
        self.selected.emit()

        return super().mouseReleaseEvent(event)


class ObjectViewer(QWidget):
    def __init__(self, parent: Optional[QWidget], icon: Optional[ObjectButton] = None):
        super().__init__(parent)
        self.icon = icon
        self.layout_ = QVBoxLayout()
        self.setLayout(self.layout_)
        self.setWhatsThis(
            "<b>Current Object</b><br/>"
            "Shows the currently selected object and its name. It can be placed by "
            "clicking the middle mouse button anywhere in the level."
        )

    @property
    def icon(self) -> Optional[ObjectButton]:
        return self._icon

    @icon.setter
    def icon(self, icon: Optional[ObjectButton]):
        self._icon = icon
        if icon is not None:
            clear_layout(self.layout_)
            icon.setFixedSize(QSize(64, 64))
            name = QLabel(icon.item.name)
            name.setFixedWidth(64)
            name.setWordWrap(True)
            name.setAlignment(Qt.AlignCenter)
            name.setContentsMargins(0, 0, 0, 0)
            self.layout_.addWidget(icon)
            self.layout_.addWidget(name)



from typing import Optional

from attr import attrs

from foundry.core.warnings.Warning import PydanticWarning, Warning
from foundry.game.gfx.objects.LevelObject import LevelObject


@attrs(slots=True, frozen=True, auto_attribs=True)
class InvalidSizeWarning(Warning):
    """
    A warning for an object exceeding its valid size range.
    """

    max_width: Optional[int] = None
    min_width: Optional[int] = None
    max_height: Optional[int] = None
    min_height: Optional[int] = None

    def check_object(self, obj: LevelObject, *args, **kwargs) -> bool:
        """
        Determines if the object should emit a warning for having an invalid size.

        Parameters
        ----------
        obj : ObjectLike
            The object to check.

        Returns
        -------
        bool
            If the object should emit a warning.
        """
        return (
            self.max_width is not None
            and obj.rendered_width > self.max_width
            or self.min_width is not None
            and obj.rendered_width < self.min_width
            or self.max_height is not None
            and obj.rendered_height > self.max_height
            or self.min_height is not None
            and obj.rendered_height < self.min_height
        )

    def get_message(self, obj: LevelObject) -> str:
        if self.max_width is not None and obj.rendered_width > self.max_width:
            return f"{obj.name} width of {obj.rendered_width} is more than its safe maximum of {self.max_width}."
        if self.min_width is not None and obj.rendered_width < self.min_width:
            return f"{obj.name} width of {obj.rendered_width} is less than its safe minimum of {self.min_width}."
        if self.max_height is not None and obj.rendered_height > self.max_height:
            return f"{obj.name} height of {obj.rendered_height} is more than its safe maximum of {self.max_height}."
        if self.min_height is not None and obj.rendered_height < self.min_height:
            return f"{obj.name} height of {obj.rendered_height} is less than its safe maximum of {self.min_height}."
        raise NotImplementedError


class PydanticInvalidSizeWarning(PydanticWarning):
    """
    A JSON model of a warning that checks for an object being in a size range.
    """

    max_width: Optional[int] = None
    min_width: Optional[int] = None
    max_height: Optional[int] = None
    min_height: Optional[int] = None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['foundry',
 'foundry.core',
 'foundry.core.player_animations',
 'foundry.core.sprites',
 'foundry.core.warnings',
 'foundry.game',
 'foundry.game.gfx',
 'foundry.game.gfx.drawable',
 'foundry.game.gfx.objects',
 'foundry.game.level',
 'foundry.gui',
 'foundry.smb3parse',
 'foundry.smb3parse.levels',
 'foundry.smb3parse.objects',
 'foundry.smb3parse.util']

package_data = \
{'': ['*'], 'foundry': ['data/*', 'data/icons/*']}

install_requires = \
['PySide6>=6.2.0,<7.0.0',
 'attrs>=21.2.0,<22.0.0',
 'pretty-errors>=1.2.24,<2.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'qt-material>=2.8.13,<3.0.0',
 'single-source>=0.2.0,<0.3.0']

entry_points = \
{'console_scripts': ['foundry = foundry.main:start']}

setup_kwargs = {
    'name': 'foundry-smb3',
    'version': '0.4.8',
    'description': 'The future of SMB3',
    'long_description': '# Foundry\n\n<p align="center">\n<a href="https://github.com/TheJoeSmo/Foundry/actions"><img alt="Actions Status" src="https://github.com/TheJoeSmo/Foundry/actions/workflows/tests.yml/badge.svg"></a>\n<a href="https://github.com/TheJoeSmo/Foundry/actions"><img alt="Actions Status" src="https://github.com/TheJoeSmo/Foundry/actions/workflows/github_pages.yml/badge.svg"></a>\n<a href="https://github.com/TheJoeSmo/Foundry/block/main/LICENSE.md"><img alt="License GPL3" src="https://img.shields.io/badge/License-GPLv3-blue.svg"></a>\n<a href="https://pypi.org/project/foundry-smb3/"><img alt="PyPI" src="https://img.shields.io/pypi/v/foundry-smb3"></a>\n<a href="https://pepy.tech/project/black"><img alt="Downloads" src="https://pepy.tech/badge/foundry-smb3"></a>\n<a href="https://github.com/psf/black"><img alt="Code style: black" src="https://img.shields.io/badge/code%20style-black-000000.svg"></a>\n</p>\n\nFoundry is the leading editor for SMB3.\n- Website: https://thejoesmo.github.io/Foundry/\n- Discord: https://discord.gg/pm87gm7\n- Documentation: https://thejoesmo.github.io/Foundry/\n- Manual: https://github.com/TheJoeSmo/Foundry/blob/master/MANUAL.md\n- Source Code: https://github.com/TheJoeSmo/Foundry\n- Bug Reporting: https://github.com/TheJoeSmo/Foundry/issues\n\nIt provides:\n- A powerful level editor\n\nTesting:\nFoundry uses `Poetry` as its package manager.  WIP.\n\nCall for Contributions\n----------------------\nFoundry is a community driven initiative that relies on your help and expertise.\n\nSmall improvements or fixes are critical to this repository\'s success.  Issues labeled `good first issue` are a great place to start.  For larger contributions WIP.\n\nYou do not need to be literate with programming to aid Foundry on its journey.  We also need help with:\n- Developing tutorials\n- Creating graphics for our brand and promotional material\n- Translation\n- Outreach and onboarding new contributors\n- Reviewing issues and suggestions\n\nIf you are undecided on where to start, we encourage you to reach out.  You can ask on our Discord or privately through email.\n\nIf you are new to open source projects and want to be caught up to speed, we recommend [this guide](https://opensource.guide/how-to-contribute/)',
    'author': 'TheJoeSmo',
    'author_email': 'joesmo.joesmo12@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9.0,<3.10',
}


setup(**setup_kwargs)



from typing import List, Tuple

from PySide6.QtCore import QEvent, QRect, Qt, Signal, SignalInstance
from PySide6.QtGui import QCursor, QFocusEvent
from PySide6.QtWidgets import QLabel, QVBoxLayout, QWidget

from foundry.game.gfx.objects.LevelObject import LevelObject
from foundry.game.level.LevelRef import LevelRef
from foundry.gui.LevelView import LevelView
from foundry.gui.ObjectList import ObjectList
from foundry.gui.util import clear_layout


class WarningList(QWidget):
    warnings_updated: SignalInstance = Signal(bool)

    def __init__(self, parent, level_ref: LevelRef, level_view_ref: LevelView, object_list_ref: ObjectList):
        super(WarningList, self).__init__(parent)

        self.level_ref = level_ref
        self.level_ref.data_changed.connect(self._update_warnings)

        self.level_view_ref = level_view_ref
        self.object_list = object_list_ref

        self.setLayout(QVBoxLayout())
        self.setWindowFlag(Qt.Popup)
        self.layout().setContentsMargins(5, 5, 5, 5)

        self._enemy_dict = {}

        self.warnings: List[Tuple[str, List[LevelObject]]] = []

    def _update_warnings(self):
        self.warnings.clear()

        level = self.level_ref.level

        for index, obj in enumerate(level.objects + level.enemies + level.jumps):
            for warning in obj.definition.get_warnings():
                if warning.check_object(obj, level=level, index=index):
                    self.warnings.append((warning.get_message(obj), [obj]))

        self.update()
        self.warnings_updated.emit(bool(self.warnings))

    def update(self):
        self.hide()

        clear_layout(self.layout())

        for warning in self.warnings:
            warning_message, related_objects = warning

            label = WarningLabel(warning_message, related_objects)
            label.hovered.connect(self._focus_objects)

            self.layout().addWidget(label)

        super(WarningList, self).update()

    def show(self):
        pos = QCursor.pos()
        pos.setY(pos.y() + 10)

        self.setGeometry(QRect(pos, self.layout().sizeHint()))

        super(WarningList, self).show()

    def _focus_objects(self):
        objects = self.sender().related_objects

        if objects:
            self.level_ref.blockSignals(True)

            self.level_view_ref.select_objects(objects)
            self.level_view_ref.scroll_to_objects(objects)
            self.object_list.update_content()

            self.level_ref.blockSignals(False)

    def focusOutEvent(self, event: QFocusEvent):
        self.hide()

        super(WarningList, self).focusOutEvent(event)


class WarningLabel(QLabel):
    hovered: SignalInstance = Signal()

    def __init__(self, text: str, related_objects: List[LevelObject]):
        super(WarningLabel, self).__init__(text)

        self.related_objects = related_objects

    def enterEvent(self, event: QEvent):
        self.hovered.emit()

        return super(WarningLabel, self).enterEvent(event)



from itertools import product
from typing import Optional, Union

from PySide6.QtCore import Signal, SignalInstance
from PySide6.QtWidgets import QWidget

from foundry.game.gfx.objects.EnemyItem import EnemyObject
from foundry.game.gfx.objects.EnemyItemFactory import EnemyItemFactory
from foundry.game.gfx.objects.LevelObject import LevelObject
from foundry.game.gfx.objects.LevelObjectFactory import LevelObjectFactory
from foundry.gui.FlowLayout import FlowLayout
from foundry.gui.ObjectIcon import ObjectButton
from foundry.gui.util import clear_layout
from foundry.smb3parse.objects import MAX_DOMAIN, MAX_ENEMY_ITEM_ID, MAX_ID_VALUE


class ObjectToolBox(QWidget):
    object_icon_clicked: SignalInstance = Signal(ObjectButton)  # type: ignore
    object_placed: SignalInstance = Signal(ObjectButton)  # type: ignore
    object_selected: SignalInstance = Signal(int)  # type: ignore

    def __init__(self, parent: Optional[QWidget] = None):
        super(ObjectToolBox, self).__init__(parent)
        self._layout = FlowLayout(self)
        self.objects: list[Union[EnemyObject, LevelObject]] = []

    def update(self):
        self._layout.clear()
        for idx, object in enumerate(self.objects):
            icon = ObjectButton(self, object)
            icon.selected.connect(self._on_icon_clicked)
            icon.selected.connect(lambda idx=idx: self.object_selected.emit(idx))
            icon.object_created.connect(lambda icon=icon: self.object_placed.emit(icon))
            self._layout.addWidget(icon)

    def add_object(self, level_object: Union[EnemyObject, LevelObject]):
        self.objects.append(level_object)
        self.update()

    def add_from_object_set(self, object_set_index: int, graphic_set_index: int = -1):
        if graphic_set_index == -1:
            graphic_set_index = object_set_index

        factory = LevelObjectFactory(
            object_set_index, graphic_set_index, 0, [], vertical_level=False, size_minimal=True
        )

        object_ids = list(range(0x00, 0x10)) + list(range(0x10, MAX_ID_VALUE, 0x10))

        for domain, obj_index in product(range(MAX_DOMAIN + 1), object_ids):
            level_object = factory.from_properties(
                domain=domain, object_index=obj_index, x=0, y=0, length=None, index=0
            )

            if not isinstance(level_object, LevelObject) or level_object.name in ["MSG_NOTHING", "MSG_CRASH"]:
                continue

            self.objects.append(level_object)
        self.update()

    def add_from_enemy_set(self, object_set_index: int):
        self.clear()
        factory = EnemyItemFactory(object_set_index, 0)

        for obj_index in range(MAX_ENEMY_ITEM_ID + 1):
            enemy_item = factory.from_properties(obj_index, x=0, y=0)

            if enemy_item.name in ["MSG_NOTHING", "MSG_CRASH"]:
                continue

            self.objects.append(enemy_item)
        self.update()

    def clear(self):
        clear_layout(self._layout)
        self._layout.clear()
        self.objects = []

    def _on_icon_clicked(self):
        self.object_icon_clicked.emit(self.sender())

    @property
    def draw_background_color(self):
        return self.objects[0].draw_background_color

    @draw_background_color.setter
    def draw_background_color(self, value):
        for index in range(self._layout.count()):
            self.objects[index].draw_background_color = value

    def place_at_front(self, object: Union[LevelObject, EnemyObject]):
        for idx, obj in enumerate(self.objects):
            if obj.name == object.name:
                self.objects.pop(idx)
                break
        self.objects.insert(0, object)
        clear_layout(self._layout)
        self.update()



from foundry.core.warnings.Warning import PydanticWarning, Warning
from foundry.game.gfx.objects.ObjectLike import ObjectLike


class InvalidObjectWarning(Warning):
    """
    A warning for an object that is invalid.
    """

    def get_message(self, obj: ObjectLike, *args, **kwargs) -> str:
        return f"Object at {obj.position.x}, {obj.position.y} will likely cause the game to crash"


class PydanticInvalidObjectWarning(PydanticWarning):
    """
    A JSON model of a warning that notifies the user of an invalid object.
    """



from itertools import product
from typing import Tuple

from PySide6.QtCore import QPoint, QRect
from PySide6.QtGui import QBrush, QColor, QImage, QPainter, QPen, Qt

from foundry import data_dir
from foundry.game.File import ROM
from foundry.game.gfx.drawable import apply_selection_overlay
from foundry.game.gfx.drawable.Block import Block
from foundry.game.gfx.GraphicsSet import GraphicsSet
from foundry.game.gfx.objects.EnemyItem import MASK_COLOR, EnemyObject
from foundry.game.gfx.objects.LevelObject import GROUND, SCREEN_HEIGHT, SCREEN_WIDTH
from foundry.game.gfx.objects.ObjectLike import (
    EXPANDS_BOTH,
    EXPANDS_HORIZ,
    EXPANDS_VERT,
)
from foundry.game.gfx.Palette import NESPalette, PaletteGroup
from foundry.game.level.Level import Level
from foundry.gui.AutoScrollDrawer import AutoScrollDrawer
from foundry.gui.settings import SETTINGS
from foundry.smb3parse.constants import OBJ_AUTOSCROLL
from foundry.smb3parse.levels import LEVEL_MAX_LENGTH
from foundry.smb3parse.objects.object_set import (
    CLOUDY_OBJECT_SET,
    DESERT_OBJECT_SET,
    DUNGEON_OBJECT_SET,
    ICE_OBJECT_SET,
)

png = QImage(str(data_dir / "gfx.png"))
png.convertTo(QImage.Format_RGB888)


def _make_image_selected(image: QImage) -> QImage:
    alpha_mask = image.createAlphaMask()
    alpha_mask.invertPixels()

    selected_image = QImage(image)

    apply_selection_overlay(selected_image, alpha_mask)

    return selected_image


def _load_from_png(x: int, y: int):
    image = png.copy(QRect(x * 16, y * 16, 16, 16))
    mask = image.createMaskFromColor(QColor(*MASK_COLOR).rgb(), Qt.MaskOutColor)
    image.setAlphaChannel(mask)

    return image


FIRE_FLOWER = _load_from_png(16, 53)
LEAF = _load_from_png(17, 53)
NORMAL_STAR = _load_from_png(18, 53)
CONTINUOUS_STAR = _load_from_png(19, 53)
MULTI_COIN = _load_from_png(20, 53)
ONE_UP = _load_from_png(21, 53)
COIN = _load_from_png(22, 53)
VINE = _load_from_png(23, 53)
P_SWITCH = _load_from_png(24, 53)
SILVER_COIN = _load_from_png(25, 53)
INVISIBLE_COIN = _load_from_png(26, 53)
INVISIBLE_1_UP = _load_from_png(27, 53)

NO_JUMP = _load_from_png(32, 53)
UP_ARROW = _load_from_png(33, 53)
DOWN_ARROW = _load_from_png(34, 53)
LEFT_ARROW = _load_from_png(35, 53)
RIGHT_ARROW = _load_from_png(36, 53)

ITEM_ARROW = _load_from_png(53, 53)

EMPTY_IMAGE = _load_from_png(0, 53)


SPECIAL_BACKGROUND_OBJECTS = [
    "blue background",
    "starry background",
    "underground background under this",
    "sets background to actual background color",
]


def _block_from_index(block_index: int, level: Level) -> Block:
    """
    Returns the block at the given index, from the TSA table for the given level.

    :param block_index:
    :param level:
    :return:
    """

    palette_group = PaletteGroup.from_tileset(level.object_set_number, level.header.object_palette_index)
    graphics_set = GraphicsSet.from_tileset(level.header.graphic_set_index)
    tsa_data = ROM().get_tsa_data(level.object_set_number)

    return Block(block_index, palette_group, graphics_set, tsa_data)


class LevelDrawer:
    def __init__(self):
        self.draw_jumps = False
        self.draw_grid = False
        self.draw_expansions = False
        self.draw_mario = False
        self.draw_jumps_on_objects = True
        self.draw_items_in_blocks = True
        self.draw_invisible_items = True
        self.draw_autoscroll = True
        self.transparency = False

        self.block_length = Block.WIDTH

        self.grid_pen = QPen(QColor(0x80, 0x80, 0x80, 0x80))
        self.grid_pen.setWidth(1)
        self.screen_pen = QPen(QColor(0xFF, 0x00, 0x00, 0xFF))
        self.screen_pen.setWidth(1)

    def draw(self, painter: QPainter, level: Level):
        self._draw_background(painter, level)

        if level.object_set_number == DESERT_OBJECT_SET:
            self._draw_desert_default_graphics(painter, level)
        elif level.object_set_number == DUNGEON_OBJECT_SET:
            self._draw_dungeon_default_graphics(painter, level)
        elif level.object_set_number == ICE_OBJECT_SET:
            self._draw_ice_default_graphics(painter, level)

        # painter.setPen(QPen(QColor(0x00, 0x00, 0x00, 0x80), width=1))
        # painter.setBrush(Qt.NoBrush)

        self._draw_objects(painter, level)

        self._draw_overlays(painter, level)

        if self.draw_expansions:
            self._draw_expansions(painter, level)

        if self.draw_mario:
            self._draw_mario(painter, level)

        if self.draw_jumps:
            self._draw_jumps(painter, level)

        if self.draw_grid:
            self._draw_grid(painter, level)

        if self.draw_autoscroll:
            self._draw_auto_scroll(painter, level)

    def _draw_background(self, painter: QPainter, level: Level):
        painter.save()

        if level.object_set_number == CLOUDY_OBJECT_SET:
            bg_color = NESPalette[
                PaletteGroup.from_tileset(level.object_set_number, level.header.object_palette_index)[3][2]
            ]
        else:
            bg_color = PaletteGroup.from_tileset(
                level.object_set_number, level.header.object_palette_index
            ).background_color

        painter.fillRect(level.get_rect(self.block_length), bg_color)

        painter.restore()

    def _draw_dungeon_default_graphics(self, painter: QPainter, level: Level):
        # draw_background
        bg_block = _block_from_index(140, level)

        for x, y in product(range(level.width), range(level.height)):
            bg_block.draw(painter, x * self.block_length, y * self.block_length, self.block_length)

        # draw ceiling
        ceiling_block = _block_from_index(139, level)

        for x in range(level.width):
            ceiling_block.draw(painter, x * self.block_length, 0, self.block_length)

        # draw floor
        upper_floor_blocks = [_block_from_index(20, level), _block_from_index(21, level)]
        lower_floor_blocks = [_block_from_index(22, level), _block_from_index(23, level)]

        upper_y = (GROUND - 2) * self.block_length
        lower_y = (GROUND - 1) * self.block_length

        for block_x in range(level.width):
            pixel_x = block_x * self.block_length

            upper_floor_blocks[block_x % 2].draw(painter, pixel_x, upper_y, self.block_length)
            lower_floor_blocks[block_x % 2].draw(painter, pixel_x, lower_y, self.block_length)

    def _draw_desert_default_graphics(self, painter: QPainter, level: Level):
        floor_level = (GROUND - 1) * self.block_length
        floor_block_index = 86

        floor_block = _block_from_index(floor_block_index, level)

        for x in range(level.width):
            floor_block.draw(painter, x * self.block_length, floor_level, self.block_length)

    def _draw_ice_default_graphics(self, painter: QPainter, level: Level):
        bg_block = _block_from_index(0x80, level)

        for x, y in product(range(level.width), range(level.height)):
            bg_block.draw(painter, x * self.block_length, y * self.block_length, self.block_length)

    def _draw_objects(self, painter: QPainter, level: Level):
        bg_palette_group = PaletteGroup.from_tileset(level.object_set_number, level.header.object_palette_index)
        spr_palette_group = PaletteGroup.from_tileset(level.object_set_number, 8 + level.header.enemy_palette_index)
        for level_object in level.objects:
            level_object.palette_group = bg_palette_group
        for enemy in level.enemies:
            enemy.palette_group = spr_palette_group

        for level_object in level.get_all_objects():

            level_object.render()

            if level_object.name.lower() in SPECIAL_BACKGROUND_OBJECTS:
                width = LEVEL_MAX_LENGTH
                height = GROUND - level_object.position.y

                blocks_to_draw = [level_object.blocks[0]] * width * height

                for index, block_index in enumerate(blocks_to_draw):
                    x = level_object.position.x + index % width
                    y = level_object.position.y + index // width

                    level_object._draw_block(painter, block_index, x, y, self.block_length, False)
            else:
                level_object.draw(painter, self.block_length, self.transparency)

            if level_object.selected:
                painter.save()

                pen = QPen(QColor(0x00, 0x00, 0x00, 0x80))
                pen.setWidth(1)
                painter.setPen(pen)
                painter.drawRect(level_object.get_rect(self.block_length))

                painter.restore()

    def _draw_overlays(self, painter: QPainter, level: Level):
        painter.save()

        for level_object in level.get_all_objects():
            name = level_object.name.lower()

            # only handle this specific enemy item for now
            if isinstance(level_object, EnemyObject) and "invisible door" not in name:
                continue

            pos = level_object.get_rect(self.block_length).topLeft()
            rect = level_object.get_rect(self.block_length)

            # invisible coins, for example, expand and need to have multiple overlays drawn onto them
            # set true by default, since for most overlays it doesn't matter
            fill_object = True

            # pipe entries
            if "pipe" in name and "can go" in name:
                if not self.draw_jumps_on_objects:
                    continue

                fill_object = False

                # center() is one pixel off for some reason
                pos = rect.topLeft() + QPoint(*(rect.size() / 2).toTuple())

                trigger_position = (level_object.position.x, level_object.position.y)

                if "left" in name:
                    image = LEFT_ARROW

                    pos.setX(rect.right())
                    pos.setY(pos.y() - self.block_length / 2)

                    # leftward pipes trigger on the column to the left of the opening
                    x, y = level_object.get_rect().bottomRight().toTuple()
                    trigger_position = (x - 1, y)

                elif "right" in name:
                    image = RIGHT_ARROW
                    pos.setX(rect.left() - self.block_length)
                    pos.setY(pos.y() - self.block_length / 2)

                elif "down" in name:
                    image = DOWN_ARROW

                    pos.setX(pos.x() - self.block_length / 2)
                    pos.setY(rect.top() - self.block_length)
                else:
                    # upwards pipe
                    image = UP_ARROW

                    pos.setX(pos.x() - self.block_length / 2)
                    pos.setY(rect.bottom())

                    # upwards pipes trigger on the second to last row
                    x, y = level_object.get_rect().bottomLeft().toTuple()
                    trigger_position = (x, y - 1)

                if not self._object_in_jump_area(level, trigger_position):
                    image = NO_JUMP

            elif "door" == name or "door (can go" in name or "invisible door" in name or "red invisible note" in name:
                fill_object = False

                if "note" in name:
                    image = UP_ARROW
                else:
                    # door
                    image = DOWN_ARROW

                pos.setY(rect.top() - self.block_length)

                x, y = level_object.position.x, level_object.position.y

                # jumps seemingly trigger on the bottom block
                if not self._object_in_jump_area(level, (x, y + 1)):
                    image = NO_JUMP

            # "?" - blocks, note blocks, wooden blocks and bricks
            elif "'?' with" in name or "brick with" in name or "bricks with" in name or "block with" in name:
                if not self.draw_items_in_blocks:
                    continue

                pos.setY(pos.y() - self.block_length)

                if "flower" in name:
                    image = FIRE_FLOWER
                elif "leaf" in name:
                    image = LEAF
                elif "continuous star" in name:
                    image = CONTINUOUS_STAR
                elif "star" in name:
                    image = NORMAL_STAR
                elif "multi-coin" in name:
                    image = MULTI_COIN
                elif "coin" in name:
                    image = COIN
                elif "1-up" in name:
                    image = ONE_UP
                elif "vine" in name:
                    image = VINE
                elif "p-switch" in name:
                    image = P_SWITCH
                else:
                    image = EMPTY_IMAGE

                # draw little arrow for the offset item overlay
                arrow_pos = QPoint(pos)
                arrow_pos.setY(arrow_pos.y() + self.block_length / 4)
                painter.drawImage(arrow_pos, ITEM_ARROW.scaled(self.block_length, self.block_length))

            elif "invisible" in name:
                if not self.draw_invisible_items:
                    continue

                if "coin" in name:
                    image = INVISIBLE_COIN
                elif "1-up" in name:
                    image = INVISIBLE_1_UP
                else:
                    image = EMPTY_IMAGE

            elif "silver coins" in name:
                if not self.draw_invisible_items:
                    continue

                image = SILVER_COIN
            else:
                continue

            if fill_object:
                for x in range(level_object.rendered_width):
                    adapted_pos = QPoint(pos)
                    adapted_pos.setX(pos.x() + x * self.block_length)

                    image = image.scaled(self.block_length, self.block_length)
                    painter.drawImage(adapted_pos, image)

                    if level_object.selected:
                        painter.drawImage(adapted_pos, _make_image_selected(image))

            else:
                image = image.scaled(self.block_length, self.block_length)
                painter.drawImage(pos, image)

        painter.restore()

    @staticmethod
    def _object_in_jump_area(level: Level, pos: Tuple[int, int]):
        for jump in level.jumps:
            jump_rect = jump.get_rect(1, level.is_vertical)

            if jump_rect.contains(QPoint(*pos)):
                return True
        else:
            return False

    def _draw_expansions(self, painter: QPainter, level: Level):
        for level_object in level.get_all_objects():
            if level_object.selected:
                painter.drawRect(level_object.get_rect(self.block_length))

            if self.draw_expansions:
                painter.save()

                painter.setPen(Qt.NoPen)

                if level_object.expands() == EXPANDS_BOTH:
                    painter.setBrush(QColor(0xFF, 0, 0xFF, 0x80))
                elif level_object.expands() == EXPANDS_HORIZ:
                    painter.setBrush(QColor(0xFF, 0, 0, 0x80))
                elif level_object.expands() == EXPANDS_VERT:
                    painter.setBrush(QColor(0, 0, 0xFF, 0x80))

                painter.drawRect(level_object.get_rect(self.block_length))

                painter.restore()

    def _draw_mario(self, painter: QPainter, level: Level):
        mario_actions = QImage(str(data_dir / "mario.png"))

        mario_actions.convertTo(QImage.Format_RGBA8888)

        mario_position = QPoint(*level.header.mario_position()) * self.block_length

        x_offset = 32 * level.start_action
        MARIO_POWERUP_Y_OFFSETS = [0, 0x20, 0x60, 0x40, 0xC0, 0xA0, 0x80, 0x60, 0xC0]
        y_offset = MARIO_POWERUP_Y_OFFSETS[SETTINGS["default_powerup"]]

        mario_cutout = mario_actions.copy(QRect(x_offset, y_offset, 32, 32)).scaled(
            2 * self.block_length, 2 * self.block_length
        )

        painter.drawImage(mario_position, mario_cutout)

    def _draw_jumps(self, painter: QPainter, level: Level):
        for jump in level.jumps:
            painter.setBrush(QBrush(QColor(0xFF, 0x00, 0x00), Qt.FDiagPattern))

            painter.drawRect(jump.get_rect(self.block_length, level.is_vertical))

    def _draw_grid(self, painter: QPainter, level: Level):
        panel_width, panel_height = level.get_rect(self.block_length).size().toTuple()

        painter.setPen(self.grid_pen)

        for x in range(0, panel_width, self.block_length):
            painter.drawLine(x, 0, x, panel_height)
        for y in range(0, panel_height, self.block_length):
            painter.drawLine(0, y, panel_width, y)

        painter.setPen(self.screen_pen)

        if level.is_vertical:
            for y in range(0, panel_height, self.block_length * SCREEN_HEIGHT):
                painter.drawLine(0, self.block_length + y, panel_width, self.block_length + y)
        else:
            for x in range(0, panel_width, self.block_length * SCREEN_WIDTH):
                painter.drawLine(x, 0, x, panel_height)

    def _draw_auto_scroll(self, painter: QPainter, level: Level):
        for item in level.enemies:
            if item.obj_index == OBJ_AUTOSCROLL:
                break
        else:
            return

        drawer = AutoScrollDrawer(item.position.y, level)

        drawer.draw(painter, self.block_length)



from typing import Union

from PySide6.QtWidgets import QMainWindow, QStatusBar

from foundry.game.gfx.objects.EnemyItem import EnemyObject
from foundry.game.gfx.objects.LevelObject import LevelObject
from foundry.game.level.LevelRef import LevelRef


class ObjectStatusBar(QStatusBar):
    def __init__(self, parent: QMainWindow, level_ref: LevelRef):
        super(ObjectStatusBar, self).__init__(parent=parent)

        self.level_ref = level_ref
        self.level_ref.data_changed.connect(self.update)

    def clear(self):
        self.clearMessage()

    def update(self):
        selected_objects = self.level_ref.selected_objects

        if selected_objects:
            self._fill(selected_objects[-1])

    def _fill(self, obj: Union[LevelObject, EnemyObject]):
        info = obj.get_status_info()

        message_parts = []

        for key, value in info:
            message_parts.append(f"{key}: {value}")

        self.showMessage(" | ".join(message_parts))



from pydantic import BaseModel

from foundry.core.warnings.Warning import Warning
from foundry.core.warnings.WarningCreator import WarningCreator


class Definition(BaseModel):
    description: str = ""
    warnings: list[WarningCreator] = []

    def get_warnings(self) -> list[Warning]:
        return self.warnings.copy()  # type: ignore

    class Config:
        use_enum_values = True



from functools import lru_cache

from PySide6.QtGui import QImage

from foundry.game.gfx.drawable import MASK_COLOR, bit_reverse
from foundry.game.gfx.GraphicsSet import GraphicsSetProtocol
from foundry.game.gfx.Palette import NESPalette, PaletteGroup

PIXEL_OFFSET = 8  # both bits describing the color of a pixel are in separate 8 byte chunks at the same index

BACKGROUND_COLOR_INDEX = 0


@lru_cache(2 ** 10)
class Tile:
    SIDE_LENGTH = 8  # pixel
    WIDTH = SIDE_LENGTH
    HEIGHT = SIDE_LENGTH

    PIXEL_COUNT = WIDTH * HEIGHT
    SIZE = 2 * PIXEL_COUNT // 8  # 1 pixel is defined by 2 bits

    def __init__(
        self,
        object_index: int,
        palette_group: PaletteGroup,
        palette_index: int,
        graphics_set: GraphicsSetProtocol,
        mirrored=False,
    ):
        start = object_index * Tile.SIZE

        self.cached_tiles = dict()

        self.palette = palette_group[palette_index]

        self.data = bytearray()
        self.pixels = bytearray()
        self.mask_pixels = bytearray()

        self.data = bytearray(bytes(graphics_set))[start : start + Tile.SIZE]

        if mirrored:
            self._mirror()

        for i in range(Tile.PIXEL_COUNT):
            byte_index = i // Tile.HEIGHT
            bit_index = 2 ** (7 - (i % Tile.WIDTH))

            left_bit = right_bit = 0

            if self.data[byte_index] & bit_index:
                left_bit = 1

            if self.data[PIXEL_OFFSET + byte_index] & bit_index:
                right_bit = 1

            color_index = (right_bit << 1) | left_bit

            color = self.palette[color_index]

            # add alpha values
            if color_index == 0:
                self.pixels.extend(MASK_COLOR)
            else:
                self.pixels.extend(NESPalette[color].toTuple()[:3])

        assert len(self.pixels) == 3 * Tile.PIXEL_COUNT

    def as_image(self, tile_length=8):
        if tile_length not in self.cached_tiles.keys():
            width = height = tile_length

            image = QImage(self.pixels, self.WIDTH, self.HEIGHT, QImage.Format_RGB888)

            image = image.scaled(width, height)

            self.cached_tiles[tile_length] = image

        return self.cached_tiles[tile_length]

    def _mirror(self):
        for byte in range(len(self.data)):
            self.data[byte] = bit_reverse[self.data[byte]]



from functools import lru_cache

from PySide6.QtCore import QPoint
from PySide6.QtGui import QColor, QImage, QPainter, Qt

from foundry.game.File import ROM
from foundry.game.gfx.drawable import MASK_COLOR, apply_selection_overlay
from foundry.game.gfx.drawable.Tile import Tile
from foundry.game.gfx.GraphicsSet import GraphicsSetProtocol
from foundry.game.gfx.Palette import NESPalette, PaletteGroup

TSA_BANK_0 = 0 * 256
TSA_BANK_1 = 1 * 256
TSA_BANK_2 = 2 * 256
TSA_BANK_3 = 3 * 256


@lru_cache(2 ** 10)
def get_block(block_index: int, palette_group: PaletteGroup, graphics_set: GraphicsSetProtocol, tsa_data: bytes):
    if block_index > 0xFF:
        rom_block_index = ROM().get_byte(block_index)  # block_index is an offset into the graphic memory
        block = Block(rom_block_index, palette_group, graphics_set, tsa_data)
    else:
        block = Block(block_index, palette_group, graphics_set, tsa_data)

    return block


class Block:
    SIDE_LENGTH = 2 * Tile.SIDE_LENGTH
    WIDTH = SIDE_LENGTH
    HEIGHT = SIDE_LENGTH

    PIXEL_COUNT = WIDTH * HEIGHT

    tsa_data = bytes()

    _block_cache = {}

    def __init__(
        self,
        block_index: int,
        palette_group: PaletteGroup,
        graphics_set: GraphicsSetProtocol,
        tsa_data: bytes,
        mirrored: bool = False,
    ):
        self.index = block_index
        self.palette_group = tuple(tuple(c for c in pal) for pal in palette_group)
        self.palette_index = (block_index & 0b1100_0000) >> 6

        # can't hash list, so turn it into a string instead
        self._block_id = (block_index, str(palette_group), graphics_set)

        lu = tsa_data[TSA_BANK_0 + block_index]
        ld = tsa_data[TSA_BANK_1 + block_index]
        ru = tsa_data[TSA_BANK_2 + block_index]
        rd = tsa_data[TSA_BANK_3 + block_index]

        self.lu_tile = Tile(lu, self.palette_group, self.palette_index, graphics_set)
        self.ld_tile = Tile(ld, self.palette_group, self.palette_index, graphics_set)

        if mirrored:
            self.ru_tile = Tile(lu, self.palette_group, self.palette_index, graphics_set, mirrored=True)
            self.rd_tile = Tile(ld, self.palette_group, self.palette_index, graphics_set, mirrored=True)
        else:
            self.ru_tile = Tile(ru, self.palette_group, self.palette_index, graphics_set)
            self.rd_tile = Tile(rd, self.palette_group, self.palette_index, graphics_set)

        self.image = QImage(Block.WIDTH, Block.HEIGHT, QImage.Format_RGB888)
        painter = QPainter(self.image)

        painter.drawImage(QPoint(0, 0), self.lu_tile.as_image())
        painter.drawImage(QPoint(Tile.WIDTH, 0), self.ru_tile.as_image())
        painter.drawImage(QPoint(0, Tile.HEIGHT), self.ld_tile.as_image())
        painter.drawImage(QPoint(Tile.WIDTH, Tile.HEIGHT), self.rd_tile.as_image())

        painter.end()

        if _image_only_one_color(self.image) and self.image.pixelColor(0, 0) == QColor(*MASK_COLOR):
            self._whole_block_is_transparent = True
        else:
            self._whole_block_is_transparent = False

    @classmethod
    def clear_cache(cls):
        cls._block_cache.clear()

    def draw(self, painter: QPainter, x, y, block_length, selected=False, transparent=False):
        block_attributes = (self._block_id, block_length, selected, transparent)

        if block_attributes not in Block._block_cache:
            image = self.image.copy()

            if block_length != Block.WIDTH:
                image = image.scaled(block_length, block_length)

            # mask out the transparent pixels first
            mask = image.createMaskFromColor(QColor(*MASK_COLOR).rgb(), Qt.MaskOutColor)
            image.setAlphaChannel(mask)

            if not transparent:  # or self._whole_block_is_transparent:
                image = self._replace_transparent_with_background(image)

            if selected:
                apply_selection_overlay(image, mask)

            Block._block_cache[block_attributes] = image

        painter.drawImage(x, y, Block._block_cache[block_attributes])

    def _replace_transparent_with_background(self, image: QImage):
        # draw image on background layer, to fill transparent pixels
        background = image.copy()
        color = NESPalette[self.palette_group[self.palette_index][0]]
        background.fill(color)

        _painter = QPainter(background)
        _painter.drawImage(QPoint(), image)
        _painter.end()

        return background


def _image_only_one_color(image):
    copy = image.copy()

    copy.convertTo(QImage.Format_Indexed8)

    return copy.colorCount() == 1



from foundry.smb3parse.util import little_endian


class Rom:
    def __init__(self, rom_data: bytearray):
        self._data = rom_data

    def little_endian(self, offset: int) -> int:
        return little_endian(self._data[offset : offset + 2])

    def write_little_endian(self, offset: int, integer: int):
        right_byte = (integer & 0xFF00) >> 8
        left_byte = integer & 0x00FF

        self.write(offset, bytes([left_byte, right_byte]))

    def read(self, offset: int, length: int) -> bytearray:
        return self._data[offset : offset + length]

    def write(self, offset: int, data: bytes):
        self._data[offset : offset + len(data)] = data

    def find(self, byte: bytes, offset: int = 0) -> int:
        return self._data.find(byte, offset)

    def int(self, offset: int) -> int:
        read_bytes = self.read(offset, 1)

        return read_bytes[0]

    def save_to(self, path: str):
        with open(path, "wb") as file:
            file.write(self._data)



from foundry.core.player_animations import (
    PLAYER_FRAME_PAGE_OFFSET,
    PLAYER_FRAME_START,
    PLAYER_FRAMES,
    PLAYER_POWER_UPS,
    PLAYER_POWER_UPS_PALETTE_COUNT,
    PLAYER_POWER_UPS_PALETTES,
    PLAYER_SUIT_PAGE_OFFSET,
    SPRITES_PER_FRAME,
)
from foundry.core.player_animations.PlayerAnimation import PlayerAnimation
from foundry.core.player_animations.PlayerAnimationData import PlayerAnimationData
from foundry.game.File import ROM
from foundry.game.gfx.GraphicsSet import GraphicalPage, GraphicsSet
from foundry.game.gfx.Palette import (
    COLORS_PER_PALETTE,
    Palette,
    PaletteGroup,
    PaletteGroupProtocol,
)


def load_animations_graphic_set(animation: PlayerAnimation, power_up: int, offsets: list[int]) -> GraphicsSet:
    return GraphicsSet(
        (
            GraphicalPage(animation.offset + offsets[power_up]),
            GraphicalPage(animation.offset + offsets[power_up]),
            GraphicalPage(animation.offset + offsets[power_up]),
            GraphicalPage(animation.offset + offsets[power_up]),
        )
    )


def get_animations_palette_index(is_mario: bool, power_up: int) -> int:
    if (power_up == 1 or power_up == 0 or power_up == 3) and not is_mario:
        return 1
    elif (power_up == 1 or power_up == 0 or power_up == 3) and is_mario:
        return 0
    elif power_up == 3:
        return 4
    else:
        return power_up


def load_player_animation(
    animation: PlayerAnimation,
    palette_group: PaletteGroupProtocol,
    is_mario: bool,
    power_up: int,
    offsets: list[int],
    is_kicking: bool = False,
):
    return PlayerAnimationData(
        animation,
        load_animations_graphic_set(animation, power_up, offsets),
        palette_group,
        get_animations_palette_index(is_mario, power_up),
        is_kicking=is_kicking,
    )


def load_player_animation_data(
    animations: list[PlayerAnimation],
    palette_group: PaletteGroupProtocol,
    is_mario: bool,
    power_up: int,
    offsets: list[int],
) -> list[PlayerAnimationData]:
    animation_data: list[PlayerAnimationData] = []
    for idx, animation in enumerate(animations):
        animation_data.append(load_player_animation(animation, palette_group, is_mario, power_up, offsets, idx == 0x2D))
    return animation_data


def load_power_up_palettes() -> PaletteGroupProtocol:
    return PaletteGroup(
        [
            Palette.from_rom(PLAYER_POWER_UPS_PALETTES + address * COLORS_PER_PALETTE)
            for address in range(PLAYER_POWER_UPS_PALETTE_COUNT)
        ]
    )


def load_power_up_offsets() -> list[int]:
    return list(o for o in ROM().bulk_read(PLAYER_POWER_UPS, PLAYER_SUIT_PAGE_OFFSET))


def load_player_animations() -> list[PlayerAnimation]:
    frame_data = ROM().bulk_read(SPRITES_PER_FRAME * PLAYER_FRAMES, PLAYER_FRAME_START)
    offset_data = ROM().bulk_read(PLAYER_FRAMES, PLAYER_FRAME_PAGE_OFFSET)
    animations = []
    for i in range(PLAYER_FRAMES):
        animations.append(PlayerAnimation(bytearray(f - 1 for f in frame_data[i * 6 : (i + 1) * 6]), offset_data[i]))
    return animations


def save_player_animations_to_rom(
    power_up_offsets: bytes, palette_group: bytes, animations: bytes, page_offsets: bytes
):
    rom = ROM()
    rom.bulk_write(bytearray(animations), PLAYER_FRAME_START)
    rom.bulk_write(bytearray(page_offsets), PLAYER_FRAME_PAGE_OFFSET)
    rom.bulk_write(bytearray(palette_group), PLAYER_POWER_UPS_PALETTES)
    rom.bulk_write(bytearray(power_up_offsets), PLAYER_SUIT_PAGE_OFFSET)



from attr import attrs
from PySide6.QtCore import QPoint, QRect, Signal, SignalInstance
from PySide6.QtGui import (
    QBrush,
    QCloseEvent,
    QMouseEvent,
    QPainter,
    QPaintEvent,
    QResizeEvent,
)
from PySide6.QtWidgets import QComboBox, QLabel, QLayout, QStatusBar, QToolBar, QWidget

from foundry import icon
from foundry.core.Position import Position
from foundry.core.UndoController import UndoController
from foundry.game.File import ROM
from foundry.game.gfx.drawable.Block import Block
from foundry.game.gfx.GraphicsSet import GraphicsSet
from foundry.game.gfx.Palette import PALETTE_GROUPS_PER_OBJECT_SET, PaletteGroup
from foundry.gui.BlockEditor import BlockEditorController as BlockEditor
from foundry.gui.CustomChildWindow import CustomChildWindow
from foundry.gui.LevelSelector import OBJECT_SET_ITEMS
from foundry.gui.Spinner import Spinner


@attrs(slots=True, auto_attribs=True)
class BlockViewerModel:
    tileset: int
    palette_group: int


class BlockViewerController(CustomChildWindow):
    tile_square_assembly_changed: SignalInstance = Signal(bytearray)  # type: ignore
    tileset_changed: SignalInstance = Signal(int)  # type: ignore
    palette_group_changed: SignalInstance = Signal(int)  # type: ignore
    destroyed: SignalInstance = Signal()  # type: ignore

    def __init__(self, parent):
        super().__init__(parent, "Block Viewer")

        self.model = BlockViewerModel(0, 0)
        self.view = BlockViewerView(parent=self)
        self.undo_controller = UndoController(ROM.get_tsa_data(self.tileset))
        self.setCentralWidget(self.view)
        self.toolbar = QToolBar(self)
        self.editor = None

        self.view.block_selected.connect(self._on_block_selected)
        self.view.destroyed.connect(self.destroy)  # type: ignore

        self.undo_action = self.toolbar.addAction(icon("rotate-ccw.svg"), "Undo Action")
        self.redo_action = self.toolbar.addAction(icon("rotate-cw.svg"), "Redo Action")
        self.prev_os_action = self.toolbar.addAction(icon("arrow-left.svg"), "Previous object set")
        self.next_os_action = self.toolbar.addAction(icon("arrow-right.svg"), "Next object set")
        self.zoom_out_action = self.toolbar.addAction(icon("zoom-out.svg"), "Zoom Out")
        self.zoom_in_action = self.toolbar.addAction(icon("zoom-in.svg"), "Zoom In")

        def change_tileset(offset: int):
            self.tileset += offset

        def change_zoom(offset: int):
            self.view.zoom += offset

        def undo(*_):
            self.undo_controller.undo()
            self._update_tsa_data()

        def redo(*_):
            self.undo_controller.redo()
            self._update_tsa_data()

        self.undo_action.triggered.connect(undo)  # type: ignore
        self.undo_action.setEnabled(False)
        self.redo_action.triggered.connect(redo)  # type: ignore
        self.redo_action.setEnabled(False)
        self.prev_os_action.triggered.connect(lambda *_: change_tileset(-1))  # type: ignore
        self.next_os_action.triggered.connect(lambda *_: change_tileset(1))  # type: ignore
        self.zoom_out_action.triggered.connect(lambda *_: change_zoom(-1))  # type: ignore
        self.zoom_in_action.triggered.connect(lambda *_: change_zoom(1))  # type: ignore

        def on_combo(*_):
            self.tileset = self.bank_dropdown.currentIndex()

        self.bank_dropdown = QComboBox(parent=self.toolbar)
        self.bank_dropdown.addItems(OBJECT_SET_ITEMS)
        self.bank_dropdown.setCurrentIndex(0)
        self.bank_dropdown.currentIndexChanged.connect(on_combo)  # type: ignore
        self.tileset_changed.connect(self.bank_dropdown.setCurrentIndex)

        def on_palette(value):
            self.palette_group = value

        self.palette_group_spinner = Spinner(self, maximum=PALETTE_GROUPS_PER_OBJECT_SET - 1, base=10)
        self.palette_group_spinner.valueChanged.connect(on_palette)  # type: ignore
        self.palette_group_changed.connect(self.palette_group_spinner.setValue)

        self.toolbar.addWidget(self.bank_dropdown)
        self.toolbar.addWidget(QLabel(" Object Palette: "))
        self.toolbar.addWidget(self.palette_group_spinner)

        self.addToolBar(self.toolbar)

        self.layout().setSizeConstraint(QLayout.SetFixedSize)

        self.setStatusBar(QStatusBar(self))

    def closeEvent(self, event: QCloseEvent):
        self.toolbar.close()
        if self.editor is not None:
            self.editor.close()
        self.destroyed.emit()
        super().closeEvent(event)

    @property
    def tsa_data(self) -> bytearray:
        return self.undo_controller.state

    @tsa_data.setter
    def tsa_data(self, tsa_data: bytearray):
        self.undo_controller.do(tsa_data)
        self._update_tsa_data()

    def _update_tsa_data(self):
        ROM.write_tsa_data(self.tileset, self.tsa_data)
        Block.clear_cache()
        self.undo_action.setEnabled(self.undo_controller.can_undo)
        self.redo_action.setEnabled(self.undo_controller.can_redo)
        self.tile_square_assembly_changed.emit(self.tsa_data)
        self.view.update()

    @property
    def tileset(self) -> int:
        return self.model.tileset

    @tileset.setter
    def tileset(self, value: int):
        self.model.tileset = min(max(value, 0), 0xE)
        self.undo_controller = UndoController(ROM.get_tsa_data(self.tileset))
        self.view.object_set = self.tileset
        self.tileset_changed.emit(self.tileset)
        self.view.update()
        if self.editor is not None:
            self.editor.silent_update_tsa_data(self.tsa_data)
            self.editor.palette_group = PaletteGroup.from_tileset(self.tileset, self.palette_group)
            self.editor.graphics_set = GraphicsSet.from_tileset(self.tileset)

    @property
    def palette_group(self) -> int:
        return self.model.palette_group

    @palette_group.setter
    def palette_group(self, value: int):
        self.model.palette_group = value
        self.view.palette_group = value
        self.palette_group_changed.emit(self.palette_group)
        self.view.update()
        if self.editor is not None:
            self.editor.palette_group = PaletteGroup.from_tileset(self.tileset, value)

    def _on_block_selected(self, index: int):
        if self.editor is None:
            self.editor = BlockEditor(
                self,
                ROM.get_tsa_data(self.tileset),
                index,
                GraphicsSet.from_tileset(self.tileset),
                PaletteGroup.from_tileset(self.tileset, self.palette_group),
                index // 0x40,
            )

            def change_tsa_data(tsa_data: bytearray):
                self.tsa_data = tsa_data

            def remove_editor(*_):
                self.editor = None

            self.editor.tile_square_assembly_changed.connect(change_tsa_data)
            self.editor.destroyed.connect(remove_editor)
            self.tile_square_assembly_changed.connect(self.editor.silent_update_tsa_data)
            self.editor.show()
        else:
            self.editor.block_index = index


class BlockViewerView(QWidget):
    block_selected: SignalInstance = Signal(int)  # type: ignore

    BLOCKS = 256
    BLOCKS_PER_ROW = 16
    BLOCKS_PER_COLUMN = 16

    def __init__(self, parent, object_set=0, palette_group=0, zoom=2):
        super().__init__(parent)

        self.setMouseTracking(True)

        self.object_set = object_set
        self.palette_group = palette_group
        self.zoom = zoom

    def resizeEvent(self, event: QResizeEvent):
        self.update()

    @property
    def zoom(self) -> int:
        return self._zoom

    @zoom.setter
    def zoom(self, value: int):
        self._zoom = value
        self.setFixedSize(
            self.BLOCKS_PER_ROW * Block.WIDTH * self.zoom, self.BLOCKS_PER_COLUMN * Block.HEIGHT * self.zoom
        )

    @property
    def block_scale(self) -> int:
        return Block.WIDTH * self.zoom

    def mouseMoveEvent(self, event: QMouseEvent):
        pos = Position.from_qpoint(event.pos())
        pos.x, pos.y = pos.x // self.block_scale, pos.y // self.block_scale

        dec_index = pos.y * self.BLOCKS_PER_ROW + pos.x
        hex_index = hex(dec_index).upper().replace("X", "x")

        status_message = f"Row: {pos.y}, Column: {pos.x}, Index: {dec_index} / {hex_index}"

        self.parent().statusBar().showMessage(status_message)  # type: ignore

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:
        pos = Position.from_qpoint(event.pos())
        pos.x, pos.y = pos.x // self.block_scale, pos.y // self.block_scale

        index = pos.y * self.BLOCKS_PER_ROW + pos.x
        self.block_selected.emit(index)

    def paintEvent(self, event: QPaintEvent):
        painter = QPainter(self)

        palette = PaletteGroup.from_tileset(self.object_set, self.palette_group)
        bg_color = palette.background_color
        painter.setBrush(QBrush(bg_color))

        painter.drawRect(QRect(QPoint(0, 0), self.size()))

        graphics_set = GraphicsSet.from_tileset(self.object_set)
        tsa_data = ROM.get_tsa_data(self.object_set)

        for i in range(self.BLOCKS):
            block = Block(i, palette, graphics_set, tsa_data)

            x = (i % self.BLOCKS_PER_ROW) * self.block_scale
            y = (i // self.BLOCKS_PER_ROW) * self.block_scale

            block.draw(painter, x, y, self.block_scale)
